from .utils import generate_prime, lf, random_lt_n
from gmpy2 import lcm, gcd, invert, powmod, mpz


class PaillierKeyPair():
    def __init__(self, public_key, private_key):
        self.public_key = public_key
        self.private_key = private_key

class PaillierPrivateKey():
    def __init__(self, l, u, n):
        self.l = l
        self.u = u
        self.n = n
        self.n_sqr = n**2

class PaillierPublicKey():
    def __init__(self, n, g):
        self.n = n
        self.g = g
        self.n_sqr = n**2

class PaillierCiphertext():
    def __init__(self, c, n):
        self.c = c
        self.n = n
        self.n_sqr = n**2

    def __add__(self, other):
        return PaillierCiphertext((self.c * other.c) % self.n_sqr, self.n)

    def __radd__(self, other):
        return __add__(self, other)

    def __mul__(self, other):
        return PaillierCiphertext((self.c ** other) % self.n_sqr, self.n)

    def __rmul__(self, other):
        return __mul__(self, other)

    def __sub__(self, other):
        return PaillierCiphertext((self.c * invert(other.c, self.n_sqr)) % self.n_sqr, self.n)

    def __rsub__(self, other):
        return __sub__(self, other)

    
def encrypt(m, public_key):
    r = random_lt_n(public_key.n)
    a = powmod(public_key.g, m, public_key.n_sqr)
    b = powmod(r, public_key.n, public_key.n_sqr)
    c = (a * b) % public_key.n_sqr
    return PaillierCiphertext(c, public_key.n)

def decrypt(p_c, private_key):
    c = p_c.c
    a = powmod(c, private_key.l, private_key.n_sqr)
    return lf(a, private_key.n) * private_key.u % private_key.n

def keygen(n_len=2048):
    n = 0
    p = q = None
    while n.bit_length() != n_len and p == q:
        p = generate_prime(n_len // 2)
        q = generate_prime(n_len // 2)
        n = p * q        

    l = lcm(p-1, q-1)
    if gcd(p*q, (p-1)*(q-1)) != 1:
        raise RuntimeException("Critical error: gcd(N,tot(N)) != 1")

    g = mpz()
    n_sqr = n**2
    while gcd(g, n_sqr) != 1:
        g = random_lt_n(n)

    u = invert(lf(powmod(g, l, n_sqr), n), n)
    public_key = PaillierPublicKey(n, g)
    private_key = PaillierPrivateKey(l, u, n)
    return PaillierKeyPair(public_key, private_key)
    



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="paillierlib",
    version="0.0.2",
    author="Carlton Shepherd",
    author_email="carlton@linux.com",
    description="A simple implementation of the Paillier cryptosystem",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/carltonshepherd/paillier-lib",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Security :: Cryptography",
        "Intended Audience :: Science/Research",
        "Development Status :: 3 - Alpha"
    ],
    keywords='paillier,encryption,decryption,homomorphic,crypto,cryptography,security,privacy',
    install_requires=[
        "gmpy2>=2.0.0",
    ]
)



import random
import gmpy2


def generate_prime(n_bits):
    rf = random.SystemRandom()
    r = gmpy2.mpz(rf.getrandbits(n_bits))
    r = gmpy2.bit_set(r, n_bits-1)
    return gmpy2.next_prime(r)


def random_lt_n(n):
    return random.SystemRandom().randrange(1, n)


def lf(x, n):
    return (x-1) // n



from paillierlib import paillier as pal
from gmpy2 import mpz

m1 = mpz(10)
m2 = mpz(1)
m3 = mpz(0)
m4 = mpz(2000)
m5 = mpz(4000)
m6 = mpz(999)

key_pair = pal.keygen()
c1 = pal.encrypt(m1, key_pair.public_key)
c2 = pal.encrypt(m2, key_pair.public_key)
c3 = pal.encrypt(m3, key_pair.public_key)
c4 = pal.encrypt(m4, key_pair.public_key)
c5 = pal.encrypt(m5, key_pair.public_key)
c6 = pal.encrypt(m6, key_pair.public_key)
    
def test_additive():
    assert pal.decrypt(c1 + c2, key_pair.private_key) == 11
    assert pal.decrypt(c1 + c2 + c1, key_pair.private_key) == 21
    assert pal.decrypt(c1 + c2 + c1 + c2, key_pair.private_key) == 22
    assert pal.decrypt(c6 + c2, key_pair.private_key) == 1000
    assert pal.decrypt(c5 + c6 + c5, key_pair.private_key) == 8999
    assert pal.decrypt(c3 + c3 + c3 + c3 + c3, key_pair.private_key) == 0

def test_sub():
    assert pal.decrypt(c2 - c3, key_pair.private_key) == 1
    assert pal.decrypt(c2 - c3 - c3 - c3, key_pair.private_key) == 1
    assert pal.decrypt(c5 - c4 - c2, key_pair.private_key) == 1999
    assert pal.decrypt(c2 + c3 + c2 - c3 - c2, key_pair.private_key) == 1 

def test_mul():
    assert pal.decrypt(c1 * m2, key_pair.private_key) == 10
    assert pal.decrypt(c1 * m1, key_pair.private_key) == 100
    assert pal.decrypt(c1 * m1 * m1 * m1 * m2, key_pair.private_key) == 10000
    assert pal.decrypt(c5 * m3, key_pair.private_key) == 0
    # TODO: more tests
